/* Copyright Rupert Smith, 2005 to 2006, all rights reserved. */
package uk.co.thebadgerset.junit.extensions;

import junit.extensions.TestDecorator;

import junit.framework.Test;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import org.apache.log4j.Logger;

import uk.co.thebadgerset.junit.extensions.util.MathUtils;

/**
 * A Decorator that runs a test repeatedly on an increasing int parameter.
 *
 * <p><table id="crc"><caption>CRC Card</caption>
 * <tr><th> Responsibilities <th> Collaborations
 * <tr><td>
 * </table>
 *
 * @author Rupert Smith
 */
public class AsymptoticTestDecorator extends WrappedSuiteTestDecorator
{
    /** Used for logging. */
    private static final Logger log = Logger.getLogger(AsymptoticTestDecorator.class);

    /** The int parameter to start testing at. */
    //private int start;

    /** The int parameter to end testing at. */
    //private int end;

    /** The step size to increment between the start and end parameters by. */
    //private int step;

    /** The int size parameters to run the test with. */
    private int[] params;

    /** The number of times the whole test should be repeated. */
    private int repeat;

    /**
     * Creates an asymptotic test decorator that wraps a test with repeats and a set of integer 'size' paramters
     * to call the test with.
     *
     * @param test   The test to wrap.
     * @param params The integer 'size' parameters.
     * @param repeat The number of times to repeat the test.
     */
    public AsymptoticTestDecorator(WrappedSuiteTestDecorator test, int[] params, int repeat)
    {
        super(test);

        log.debug("public AsymptoticTestDecorator(Test \"" + test + "\", int[] "
                  + ((params == null) ? null : MathUtils.printArray(params)) + ", int " + repeat + "): called");

        this.params = params;
        this.repeat = repeat;
    }

    /**
     * Creates a new AsymptoticTestDecorator object.
     *
     * @param test   The test to decorate.
     * @param start  The starting asymptotic integer parameter value.
     * @param end    The ending asymptotic integer parameter value.
     * @param step   The increment size to move from the start to end values by.
     * @param repeat The number of times to repeat the test at each step of the cycle.
     */
    public AsymptoticTestDecorator(WrappedSuiteTestDecorator test, int start, int end, int step, int repeat)
    {
        super(test);

        if (start < 0)
        {
            throw new IllegalArgumentException("Start must be >= 0");
        }

        if (end < start)
        {
            throw new IllegalArgumentException("End must be >= start");
        }

        if (step < 1)
        {
            throw new IllegalArgumentException("Step must be >= 1");
        }

        if (repeat < 1)
        {
            throw new IllegalArgumentException("Repeat must be >= 1");
        }

        // Generate the sequence.
        params = new int[((end - start) / step) + 1];
        int i = 0;
        for (int n = start; n <= end; n += step)
        {
            params[i++] = n;
        }

        this.repeat = repeat;
    }

    /*
     * Works out the total number of test cases that will be run by this decorator. This is the number from
     * the start int parameter to the end int parameter multiplied by the number of repeats.
     *
     * @return The total number of test cases that will be run by this decorator.
     */
    /*public int countTestCases()
    {
        return params.length;
    }*/

    /**
     * Runs the test repeatadly for each value of the int parameter specified and for the correct number of test
     * repeats.
     *
     * @param result The test result object that the tests will indicate their results to. This is also used
     *               to pass the int parameter from this class to the decorated test class.
     */
    public void run(TestResult result)
    {
        log.debug("public void run(TestResult result): called");

        if (!(result instanceof TKTestResult))
        {
            throw new IllegalArgumentException("AsymptoticTestDecorator only works with TKTestResult");
        }

        // Cast the test result into a TKTestResult to place the current parameter into.
        TKTestResult tkResult = (TKTestResult) result;

        log.debug("params = " + ((params == null) ? null : MathUtils.printArray(params)));
        log.debug("repeat = " + repeat);

        for (int i = 0; i < params.length; i++)
        {
            int n = params[i];

            for (int j = 0; j < repeat; j++)
            {
                log.debug("n = " + n);

                // Set the integer parameter in the TKTestResult to be passed to the tests.
                tkResult.setN(n);

                if (tkResult.shouldStop())
                {
                    log.debug("tkResult.shouldStop = " + true);

                    break;
                }

                log.debug("Calling super#run");
                super.run(tkResult);
            }
        }
    }

    /**
     * Prints out the name of this test with the string "(parameterized)" appended onto it for debugging purposes.
     *
     * @return The name of this test with the string "(parameterized)" appended onto it.
     */
    public String toString()
    {
        return super.toString() + "(parameterized)";
    }
}
