/* Copyright Rupert Smith, 2005 to 2006, all rights reserved. */
package uk.co.thebadgerset.junit.extensions;

import junit.extensions.TestDecorator;

import junit.framework.Test;
import junit.framework.TestResult;

import org.apache.log4j.Logger;

/**
 * A test decorator that runs a test repeatedly until a specified length of time has passed.
 *
 * <p/><table id="crc"><caption>CRC Card</caption>
 * <tr><th> Responsibilities <th> Collaborations
 * <tr><td> Repeatedly run a test for a fixed length of time.
 * </table>
 *
 * @todo The count of the number of tests run is an important number to keep. Also num passed/error/failed is also
 *       important to record. What to do with these numbers? They are already logged to the test listeners.
 *
 * @todo The duration test runner wraps on top of size, repeat or thread wrappers, need a way for it to tell
 *       TKTestResult when the duration is up, so that it can terminate any repeats in progress. It should end
 *       as soon as possible once the test method exits.
 *
 * @author Rupert Smith
 */
public class DurationTestDecorator extends WrappedSuiteTestDecorator
{
    /** Used for logging. */
    private static final Logger log = Logger.getLogger(DurationTestDecorator.class);

    /** The test to run. */
    private Test test;

    /** The length of time to run the test for. */
    private long duration;

    /**
     * Creates an active test with default multiplier (1).
     *
     * @param test The target test.
     */
    public DurationTestDecorator(WrappedSuiteTestDecorator test)
    {
        super(test);
        this.test = test;
    }

    /**
     * Creates active test with default multiplier (1).
     *
     * @param test     The target test.
     * @param duration The duration in milliseconds.
     */
    public DurationTestDecorator(WrappedSuiteTestDecorator test, long duration)
    {
        super(test);

        log.debug("public DurationTestDecorator(Test \"" + test + "\", long " + duration + "): called");

        this.test = test;
        this.duration = duration;
    }

    /**
     * Runs the test repeatedly for the fixed duration.
     *
     * @param testResult The the results object to monitor the test results with.
     */
    public void run(TestResult testResult)
    {
        log.debug("public void run(TestResult testResult): called");

        long start = System.nanoTime();
        long now = start;
        long end = (duration * 1000000) + now;

        // Run the test until the duration times out and keep a count of how many tests ran.
        int count = 0;

        for (; now < end; count++)
        {
            test.run(testResult);

            now = System.nanoTime();
        }

        long actualEnd = System.nanoTime();
    }
}
