/*
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */
/* Copyright Rupert Smith, 2005 to 2006, all rights reserved. */
package uk.co.thebadgerset.junit.extensions;

import java.io.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Properties;

import junit.extensions.RepeatedTest;

import junit.framework.Test;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import junit.textui.ResultPrinter;
import junit.textui.TestRunner;

import org.apache.log4j.Logger;

import uk.co.thebadgerset.junit.extensions.listeners.CSVTestListener;
import uk.co.thebadgerset.junit.extensions.util.CommandLineParser;
import uk.co.thebadgerset.junit.extensions.util.MathUtils;
import uk.co.thebadgerset.junit.extensions.util.ParsedProperties;
import uk.co.thebadgerset.junit.extensions.util.TestContextProperties;

/**
 * TKTestRunner extends {@link junit.textui.TestRunner} with the ability to run tests multiple times, to execute a test
 * simultaneously using many threads, to put a delay between test runs and adds support for tests that take integer
 * parameters that can be 'stepped' through on multiple test runs. These features can be accessed by using this class
 * as an entry point and passing command line arguments to specify which features to use:
 *
 * <pre>
 * -w ms       The number of milliseconds between invocations of test cases.
 * -c pattern  The number of tests to run concurrently.
 * -r num      The number of times to repeat each test.
 * -d duration The length of time to run the tests for.
 * -t name     The name of the test case to execute.
 * -s pattern  The size parameter to run tests with.
 * -o dir      The name of the directory to output test timings to.
 * -v          Verbose mode.
 * </pre>
 *
 * <p/>The pattern arguments are of the form [lowest(, ...)(, highest)](,sample=s)(,exp), where round brackets
 * enclose optional values. Using this pattern form it is possible to specify a single value, a range of values divided
 * into s samples, a range of values divided into s samples but distributed exponentially, or a fixed set of samples.
 *
 * <p/>The duration arguments are of the form dD(:hH)(:mM)(:sS), where round brackets enclose optional values.
 *
 * <p/>Here are some examples:
 *
 * <p/><table>
 * <tr><td><pre> -c [10,20,30,40,50] </pre><td> Runs the test with 10,20,...,50 threads.
 * <tr><td><pre> -s [1,100],samples=10 </pre>
 *     <td> Runs the test with ten different size parameters evenly spaced between 1 and 100.
 * <tr><td><pre> -s [1,1000000],samples=10,exp </pre>
 *     <td> Runs the test with ten different size parameters exponentially spaced between 1 and 1000000.
 * <tr><td><pre> -r 10 </pre><td> Runs each test ten times.
 * <tr><td><pre> -d 10H </pre><td> Runs the test repeatedly for 10 hours.
 * <tr><td><pre> -d 1M, -r 10 </pre>
 *     <td> Runs the test repeatedly for 1 minute but only takes a timing sample every 10 test runs.
 * <tr><td><pre> -r 10, -c [1, 5, 10, 50], -s [100, 1000, 10000] </pre>
 *     <td> Runs 12 test cycles (4 concurrency samples * 3 size sample), with 10 repeats each. In total the test
 *          will be run 199 times (3 + 15 + 30 + 150)
 * </table>
 *
 * <p><table id="crc"><caption>CRC Card</caption>
 * <tr><th> Responsibilities <th> Collaborations
 * <tr><td> Create the test configuration specified by the command line parameters.
 * </table>
 *
 * @todo Verify that the output directory exists or can be created.
 *
 * @todo Verify that the specific named test case to execute exists.
 *
 * @todo Drop the delay parameter as it is being replaced by throttling.
 *
 * @author Rupert Smith
 */
public class TKTestRunner extends TestRunnerImprovedErrorHandling
{
    /** Used for logging. */
    private static final Logger log = Logger.getLogger(TKTestRunner.class);

    /** Used for generating the timestamp when naming output files. */
    private static final DateFormat TIME_STAMP_FORMAT = new SimpleDateFormat("yyyy-MM-dd-hh.mm.ss");

    /**
     * A copy of the result printer. This instance variable is declared private by TestRunner so a copy is placed
     * here too.
     */
    private ResultPrinter fPrinter;

    /** Enable verbose output. */
    private boolean verbose = false;

    /** Number of times to rerun the test. */
    private Integer repetitions = 1;

    /** The length of time to run the tests for. */
    private Long duration;

    /** Number of threads running the tests. */
    private int[] threads;

    /** Delay in ms to wait between two test cases. */
    private int delay = 0;

    /** The parameter values to pass to parameterized tests. */
    private int[] params;

    /** Name of the single test case to execute. */
    private String testCaseName = null;

    /** Name of the test class. */
    private String testClassName = null;

    /** Name of the test run. */
    private String testRunName = null;

    /** Directory to output XML reports into, if specified. */
    private String reportDir = null;
    private TKTestResult result;

    /**
     * Constructs a TKTestRunner using System.out for all the output.
     */
    public TKTestRunner()
    {
        this(System.out);

        log.debug("public TKTestRunner(): called");
    }

    /**
     * Constructs a TKTestRunner using the given stream for all the output.
     *
     * @param writer The print stream to send all test output to.
     */
    public TKTestRunner(PrintStream writer)
    {
        this(new ResultPrinter(writer));

        log.debug("public TKTestRunner(PrintStream writer): called");
    }

    /**
     * Constructs a TKTestRunner using the given ResultPrinter all the output.
     *
     * @param printer The ResultPrinter to send all test output to.
     */
    public TKTestRunner(ResultPrinter printer)
    {
        super(printer);

        log.debug("public TKTestRunner(ResultPrinter printer): called");

        this.fPrinter = printer;
    }

    /**
     * The entry point for the toolkit test runner.
     *
     * @param args The command line arguments.
     */
    public static void main(String[] args)
    {
        // Use the command line parser to evaluate the command line.
        CommandLineParser commandLine =
            new CommandLineParser(
                new String[][]
                {
                    { "w", "The number of milliseconds between invocations of test cases.", "ms", "false" },
                    { "c", "The number of tests to run concurrently.", "num", "false", MathUtils.SEQUENCE_REGEXP },
                    { "r", "The number of times to repeat each test.", "num", "false" },
                    { "d", "The length of time to run the tests for.", "duration", "false", MathUtils.DURATION_REGEXP },
                    { "f", "The maximum rate to call the tests at.", "frequency", "false", "^([1-9][0-9]*)/([1-9][0-9]*)$" },
                    { "s", "The size parameter to run tests with.", "size", "false", MathUtils.SEQUENCE_REGEXP },
                    { "t", "The name of the test case to execute.", "name", "false" },
                    { "o", "The name of the directory to output test timings to.", "dir", "false" },
                    { "v", "Verbose mode.", "false" },
                    { "n", "A name for this test run, used to name the output file.", "name" },
                    { "1", "Test class.", "class", "true" }
                });

        // Capture the command line arguments or display errors and correct usage and then exit.
        ParsedProperties options = null;

        try
        {
            options = new ParsedProperties(commandLine.parseCommandLine(args));
        }
        catch (IllegalArgumentException e)
        {
            System.out.println(commandLine.getErrors());
            System.out.println(commandLine.getUsage());
            System.exit(FAILURE_EXIT);
        }

        // Extract the command line options.
        Integer delay = options.getPropertyAsInteger("w");
        String threadsString = options.getProperty("c");
        Integer repetitions = options.getPropertyAsInteger("r");
        String durationString = options.getProperty("d");
        String frequencyString = null;
        String paramsString = options.getProperty("s");
        String testCaseName = options.getProperty("t");
        String reportDir = options.getProperty("o");
        boolean verbose = options.getPropertyAsBoolean("v");
        String testClassName = options.getProperty("1");
        String testRunName = options.getProperty("n");

        int[] threads = (threadsString == null) ? null : MathUtils.parseSequence(threadsString);
        int[] params = (paramsString == null) ? null : MathUtils.parseSequence(paramsString);
        Long duration = (durationString == null) ? null : (Long) MathUtils.parseDuration(durationString);

        // The test run name defaults to the test class name unless a value was specified for it.
        testRunName = (testRunName == null) ? testClassName : testRunName;

        // Add all the command line options to system properties. Tests may pick up overridden values from there.
        commandLine.addCommandLineToSysProperties();

        // Create and start the test runner.
        try
        {
            TKTestRunner testRunner = new TKTestRunner();
            TestResult testResult =
                testRunner.start(repetitions, duration, threads, (delay == null) ? 0 : delay, params, testClassName,
                                 testCaseName, reportDir, testRunName, verbose);

            if (!testResult.wasSuccessful())
            {
                System.exit(FAILURE_EXIT);
            }
            else
            {
                System.exit(SUCCESS_EXIT);
            }
        }
        catch (Exception e)
        {
            System.err.println(e.getMessage());
            System.exit(EXCEPTION_EXIT);
        }
    }

    /**
     * Runs a test or suite of tests, using the super class implemenation. This method wraps the test to be run
     * in any test decorators needed to add in the configured toolkits enhanced junit functionality.
     *
     * @param test The test to run.
     * @param wait Undocumented. Nothing in the JUnit javadocs to say what this is for.
     *
     * @return The results of the test run.
     */
    public TestResult doRun(Test test, boolean wait)
    {
        log.debug("public TestResult doRun(Test \"" + test + "\", boolean " + wait + "): called");
        log.debug("params = " + ((params == null) ? null : MathUtils.printArray(params)));
        log.debug("repetitions = " + repetitions);
        log.debug("threads = " + ((threads == null) ? null : MathUtils.printArray(threads)));
        log.debug("duration = " + duration);

        WrappedSuiteTestDecorator targetTest = null;

        if (test instanceof TestSuite)
        {
            log.debug("targetTest is a TestSuite");

            TestSuite suite = (TestSuite) test;

            int numTests = suite.countTestCases();
            log.debug("There are " + numTests + " in the suite.");

            for (int i = 0; i < numTests; i++)
            {
                Test nextTest = suite.testAt(i);
                log.debug("suite.testAt(" + i + ") = " + nextTest);

                if (nextTest instanceof TimingControllerAware)
                {
                    log.debug("nextTest is TimingControllerAware");
                }

                if (nextTest instanceof TestThreadAware)
                {
                    log.debug("nextTest is TestThreadAware");
                }
            }

            targetTest = new WrappedSuiteTestDecorator(suite);
            log.debug("Wrapped with a WrappedSuiteTestDecorator.");
        }

        // If size parameter values have been set, then wrap the test in an asymptotic test decorator.
        if (params != null)
        {
            targetTest = new AsymptoticTestDecorator(targetTest, params, (repetitions == null) ? 1 : repetitions);
            log.debug("Wrapped with asymptotic test decorator.");
            log.debug("targetTest = " + targetTest);
        }
        // If no size parameters are set but the repitions parameter is, then wrap the test in a repeated test decorator.
        else if ((repetitions != null) && (repetitions > 1))
        {
            targetTest = new AsymptoticTestDecorator(targetTest, new int[] { 1 }, repetitions);
            log.debug("Wrapped with asymptotic test decorator.");
            log.debug("targetTest = " + targetTest);
        }

        // If a test run duration has been set then wrap the test in a duration test decorator. This will wrap on
        // top of size, repeat or concurrency wrappings already applied.
        if (duration != null)
        {
            targetTest = new DurationTestDecorator(targetTest, duration);
            log.debug("Wrapped with duration test decorator.");
            log.debug("targetTest = " + targetTest);
        }

        // If a test thread concurrency level is set then wrap the test in a scaled test decorator. This will wrap on
        // top of size scaling or repetition wrappings.
        ScaledTestDecorator scaledDecorator = null;
        if ((threads != null) && ((threads.length > 1) || (MathUtils.maxInArray(threads) > 1)))
        {
            scaledDecorator = new ScaledTestDecorator(targetTest, threads);
            targetTest = scaledDecorator;
            log.debug("Wrapped with scaled test decorator.");
            log.debug("targetTest = " + targetTest);
        }
        else
        {
            scaledDecorator = new ScaledTestDecorator(targetTest, new int[] { 1 });
            targetTest = scaledDecorator;
            log.debug("Wrapped with scaled test decorator with default of 1 thread.");
            log.debug("targetTest = " + targetTest);
        }

        // Delegate to the super method to run the decorated tests.
        log.debug("About to call super.doRun");
        TestResult result = super.doRun(targetTest, wait);
        log.debug("super.doRun returned.");

        if (result instanceof TKTestResult)
        {
            TKTestResult tkResult = (TKTestResult) result;

            tkResult.notifyEndBatch();
        }

        // Shutdown the thread pool if a scaled test decorator was used.
        if (scaledDecorator != null)
        {
            scaledDecorator.shutdown();
        }

        return result;
    }

    /**
     * Creates the TestResult object to be used for test runs. See {@link TKTestResult} for more information and the
     * enhanced test result class that this uses.
     *
     * @return An instance of the enhanced test result object, {@link TKTestResult}.
     */
    protected TestResult createTestResult()
    {
        log.debug("protected TestResult createTestResult(): called");

        result = new TKTestResult(this.fPrinter.getWriter(), this.delay, this.verbose, this.testCaseName);

        // Check if a directory to output reports to has been specified and attach test listeners and timing listeners
        // if so.
        if (reportDir != null)
        {
            // Create the report directory if it does not already exist.
            File reportDirFile = new File(reportDir);

            if (!reportDirFile.exists())
            {
                reportDirFile.mkdir();
            }

            // Create the timings file (make the name of this configurable as a command line parameter).
            Writer timingsWriter = null;

            try
            {
                File timingsFile =
                    new File(reportDirFile, testRunName + "-" + TIME_STAMP_FORMAT.format(new Date()) + "-timings.csv");

                timingsWriter = new BufferedWriter(new FileWriter(timingsFile), 20000);
            }
            catch (IOException e)
            {
                log.error(e.getMessage(), e);
            }

            CSVTestListener listener = new CSVTestListener(timingsWriter);

            result.addListener(listener);
            result.addTKTestListener(listener);

            // Record the start time of the batch.
            result.notifyStartBatch();

            // At this point in time the test class has been instantiated, giving it an opportunity to read its parameters.
            // Inform any test listers of the test properties.
            result.notifyTestProperties(TestContextProperties.getAccessedProps());
        }

        return result;
    }

    /**
     * Initializes the test runner with the provided command line arguments and and starts the test run.
     *
     * @param repetitions   The number of times to repeat the test, or test batch size.
     * @param duration      The length of time to run the tests for. -1 means no duration has been set.
     * @param threads       The concurrency levels to ramp up to.
     * @param delay         A delay in milliseconds between test runs.
     * @param params        The sets of 'size' parameters to pass to test.
     * @param testClassName The fully qualified name of the test class to run.
     * @param testCaseName  The name of the test case to run.
     * @param reportDir     The directory to output the test results to.
     * @param runName       The name of the test run; used to name the output file.
     * @param verbose       Whether to print comments during test run.
     *
     * @return The test results.
     *
     * @throws Exception Any exceptions from running the tests are allowed to fall through.
     */
    protected TestResult start(Integer repetitions, Long duration, int[] threads, int delay, int[] params,
                               String testClassName, String testCaseName, String reportDir, String runName, boolean verbose)
                        throws Exception
    {
        // Keep all the test parameters.
        this.verbose = verbose;
        this.repetitions = repetitions;
        this.duration = duration;
        this.threads = threads;
        this.delay = delay;
        this.params = params;
        this.testCaseName = testCaseName;
        this.reportDir = reportDir;
        this.testClassName = testClassName;
        this.testRunName = runName;

        // Delegate to the super method to run the tests.
        return super.start(new String[] { testClassName });
    }
}
