/* Copyright Rupert Smith, 2005 to 2006, all rights reserved. */
package uk.co.thebadgerset.junit.extensions.example;

import uk.co.thebadgerset.junit.extensions.AsymptoticTestCase;
import uk.co.thebadgerset.junit.extensions.Throttle;

/**
 * ThrottleTestPerfBase is a base implementation of tests for performance testing the accuracy of throttle
 * implementations.
 *
 * <p/><table id="crc"><caption>CRC Card</caption>
 * <tr><th> Responsibilities <th> Collaborations
 * <tr><td> Check one second accuracy of throttle by running for number of ops per second.
 *     <td> {@link Throttle}
 * </table>
 *
 * @author Rupert Smith
 */
public abstract class ThrottleTestPerfBase extends AsymptoticTestCase
{
    /** Used to hold the test setup on a per thread basis. */
    ThreadLocal<Throttle> threadSetup = new ThreadLocal<Throttle>();

    /**
     * Creates a named throttle test.
     *
     * @param name The name of the test.
     */
    public ThrottleTestPerfBase(String name)
    {
        super(name);
    }

    /**
     * Should be implemented to provide a throttle implementation to test.
     *
     * @return The throttle to test.
     */
    public abstract Throttle getTestThrottle();

    /**
     * Check one second accuracy of throttle by running for number of ops per second.
     *
     * @param opsPerSecond The test size, interpreted as ops per second.
     */
    public void testThrottleAccuracy(int opsPerSecond)
    {
        // Ensure the test throttle is setup on the current test thread.
        Throttle testThrottle = threadSetup.get();
        if (testThrottle == null)
        {
            testThrottle = getTestThrottle();
            threadSetup.set(testThrottle);
        }

        // Setup the throttle to the desired rate and reset it.
        testThrottle.setRate(opsPerSecond);

        // Loop for the specified operations per second, to get a one second test.
        // The first loop returns straight away, so one extra is done.
        for (int i = 0; i <= opsPerSecond; i++)
        {
            testThrottle.throttle();
        }
    }
}
