/*
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */
/* Copyright Rupert Smith, 2005 to 2006, all rights reserved. */
package uk.co.thebadgerset.junit.extensions.example;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.apache.log4j.Logger;

import uk.co.thebadgerset.junit.extensions.AsymptoticTestCase;
import uk.co.thebadgerset.junit.extensions.util.ParsedProperties;
import uk.co.thebadgerset.junit.extensions.util.TestContextProperties;

/**
 * TimingExampleTest implements some short pauses that test invocations can try out as dummy test material.
 * The pause time per test, or per setup method can be configured by setting properties.
 *
 * <p/><table id="crc"><caption>CRC Card</caption>
 * <tr><th> Responsibilities <th> Collaborations
 * <tr><td> Pause a specified number of times.
 * </table>
 *
 * @author Rupert Smith
 */
public class TimingExampleTestPerf extends AsymptoticTestCase
{
    /** Used for logging. */
    private static final Logger log = Logger.getLogger(TimingExampleTestPerf.class);

    /** The test pause property name. */
    private static final String SLEEP_TIME_PROPNAME = "testPause";

    /** The setup pause property name. */
    private static final String SLEEP_TIME_SETUP_PROPNAME = "setupPause";

    /** The default time to sleep in the test wait loop. */
    private static final long SLEEP_TIME_DEFAULT = 10;

    /** The default time to sleep during the setup method. */
    private static final long SLEEP_TIME_SETUP_DEFAULT = 0;

    /** Used to read the tests configurable properties through. */
    ParsedProperties testProps = new TestContextProperties(System.getProperties());
    //ParsedProperties testProps = new ParsedProperties(System.getProperties());

    /** Holds the time to sleep in the test wait loop. */
    private Long testSleepTime = testProps.setPropertyIfNull(SLEEP_TIME_PROPNAME, SLEEP_TIME_DEFAULT);

    /** Holds the time to sleep during the setup method. */
    private Long setupSleepTime = testProps.setPropertyIfNull(SLEEP_TIME_SETUP_PROPNAME, SLEEP_TIME_SETUP_DEFAULT);

    /**
     * Creates a new example test case object.
     *
     * @param name The name of the test.
     */
    public TimingExampleTestPerf(String name)
    {
        super(name);
    }

    /**
     * Compiles all the tests in this class into a suite.
     *
     * @return The test suite.
     */
    public static Test suite()
    {
        // Build a new test suite
        TestSuite suite = new TestSuite("Timing Example Tests");

        suite.addTest(new TimingExampleTestPerf("testShortWait"));
        //suite.addTest(new TimingExampleTestPerf("testVeryShortWait"));

        return suite;
    }

    /**
     * Does n short waits.
     *
     * @param n The number of waits.
     */
    public void testShortWait(int n)
    {
        log.debug("public void testShortWait(int " + n + "): called");

        long sleepTime = testSleepTime;

        for (int i = 0; i < n; i++)
        {
            try
            {
                Thread.sleep(sleepTime);
            }
            catch (InterruptedException e)
            { }
        }
    }

    /**
     * Loops n times but does very little.
     *
     * @param n The number of waits.
     */
    public void testVeryShortWait(int n)
    {
        log.debug("public void testVeryShortWait(int " + n + ")");

        for (int i = 0; i < n; i++)
        {
            // Execute a few instructions, but nothing that takes very long...
            double result = Math.pow(Math.E, i);
        }
    }

    /**
     * Setup that can be configured to pause for a short time.
     */
    protected void setUp()
    {
        long sleepTime = setupSleepTime;

        try
        {
            Thread.sleep(sleepTime);
        }
        catch (InterruptedException e)
        { }
    }
}
