/* Copyright Rupert Smith, 2005 to 2006, all rights reserved. */
package uk.co.thebadgerset.junit.extensions.util;

import java.util.Properties;

/**
 * TestContextProperties is an extension of {@link ContextualProperties} that keeps track of property key/value pairs
 * that are used by tests being run under the {@link uk.co.thebadgerset.junit.extensions.TKTestRunner}. To keep the
 * test runner notified of configurable test parameters, tests should establish their required property values by
 * initiliazing fields or statics or in the constructor, through this class. The tk test runner automatically places
 * any additional properties specified on the command line into the system properties, on which an instance of this
 * class should be based.
 *
 * <p/>Here is an example:
 *
 * <pre>
 * public class MyTestClass extends TestCase {
 *     ParsedProperties testProps = new TestContextProperties(System.getProperties());
 *     private int testParam = testProps.setPropertyIfNull("testParam", 1);
 * ...
 * </pre>
 *
 * <p/>This has the effect of setting up the field testParam with the default value of 1, unless it is overridden
 * by values passed to the tk test runner. It also notifies the tk test runner of the name and value of the test
 * parameter actually used for the test, so that this can be logged in the test output file.
 *
 * <p/><table id="crc"><caption>CRC Card</caption>
 * <tr><th> Responsibilities <th> Collaborations
 * <tr><td> Log all name/value pairs read or written.
 * </table>
 *
 * @author Rupert Smith
 */
public class TestContextProperties extends ContextualProperties
{
    /** Holds all properties set or read through this property extension class. */
    static Properties accessedProps = new Properties();

    /**
     * Default constructor that builds a ContextualProperties that uses environment defaults.
     */
    public TestContextProperties()
    {
        super();
    }

    /**
     * Creates a ContextualProperties that uses environment defaults and is initialized with the specified properties.
     *
     * @param props The properties to initialize this with.
     */
    public TestContextProperties(Properties props)
    {
        super(props);
    }

    /**
     * Gets all of the properties (with their most recent values) that have been set or read through this class.
     *
     * @return All of the properties accessed through this class.
     */
    public static Properties getAccessedProps()
    {
        return accessedProps;
    }

    /**
     * Looks up a property value relative to the environment, callers class and method. The default environment will be
     * checked for a matching property if defaults are being used. The property key/value pair is remembered and made
     * available to {@link uk.co.thebadgerset.junit.extensions.TKTestRunner}.
     *
     * @param key The property key.
     *
     * @return The value of this property searching from the most specific definition (environment, class, method, key)
     *         to the most general (key only), unless use of default environments is turned off in which case the most general
     *         proeprty searched is (environment, key).
     */
    public String getProperty(String key)
    {
        String value = super.getProperty(key);
        accessedProps.setProperty(key, value);

        return value;
    }

    /**
     * Calls the <tt>Hashtable</tt> method <code>put</code>. Provided for parallelism with the <tt>getProperty</tt>
     * method. Enforces use of strings for property keys and values. The value returned is the result of the
     * <tt>Hashtable</tt> call to <code>put</code>. The property key/value pair is remembered and made
     * available to {@link uk.co.thebadgerset.junit.extensions.TKTestRunner}.
     *
     * @param key   The key to be placed into this property list.
     * @param value The value corresponding to <tt>key</tt>.
     *
     * @return The previous value of the specified key in this property list, or <code>null</code> if it did not have one.
     */
    public synchronized Object setProperty(String key, String value)
    {
        Object result = super.setProperty(key, value);
        accessedProps.setProperty(key, value);

        return result;
    }

    /**
     * Helper method for setting properties to defaults when they are not already set. The property key/value pair is
     * remembered and made available to {@link uk.co.thebadgerset.junit.extensions.TKTestRunner}.
     *
     * @param key      The name of the system property to set.
     * @param value    The value to set it to.
     *
     * @return The value of the property, which will be the value passed in if it was null, or the existing value otherwise.
     */
    public String setPropertyIfNull(String key, String value)
    {
        String result = super.setPropertyIfNull(key, value);
        accessedProps.setProperty(key, result);

        return result;
    }

}
